﻿// *********************************************************
//   NDitt Ver.0.05 -- Next-gen Ditt --                    *
//   Copyright (C) Yumitaro 2011,2020 All rights reserved. *
// *********************************************************


#include <windows.h>
#include <mmsystem.h>
#include <ctype.h>
#include <shlwapi.h>	// ie依存

#include <cstdlib>
#include <locale>
#include <string>
#include <vector>

#include "fdimg.h"
#include "fdc.h"

#define	OPT_PROTECT			0x01
#define	OPT_VERBOSE			0x02
#define	OPT_OVERTRACK		0x04
#define	OPT_D5INCH			0x08
#define	OPT_SCANSD			0x10
#define	OPT_1DDITT			0x20
#define	OPT_UNNEEDED		0x40

#define	PROTECT				(Option & OPT_PROTECT)
#define	VERBOSE				(Option & OPT_VERBOSE)
#define	OVERTRACK			(Option & OPT_OVERTRACK)
#define	D5INCH				(Option & OPT_D5INCH)
#define	SCANSD				(Option & OPT_SCANSD)
#define	F1DDITT				(Option & OPT_1DDITT)
#define	UNNEEDED			(Option & OPT_UNNEEDED)


#define	DEFAULTEXT			".d88"	// 出力ファイルのデフォルト拡張子


enum CmdType { cmdNone=0, cmdRead, cmdWrite, cmdUnformat };
enum DatRate { Rate500K=0, Rate300K, Rate250K, Rate1M };

const std::vector<std::string> comname = { "None", "Read", "Write", "Unformat" };
const std::vector<std::string> bpsname = { "500Kbps", "300Kbps", "250Kbps", "1Mbps" };


BYTE Option   = 0;			// コマンドラインオプション
BYTE SCyl     = 0;			// 開始シリンダ
BYTE ECyl     = 255;		// 終了シリンダ
BYTE DiskType = FDUNKNOWN;	// メディアタイプ

FDC fdc;




///////////////////////////////////////////////////////////////////////////////
void WriteCon( const char* format, ... )
{
	char str[1024];
	DWORD ret;
	
	va_list args;
	va_start( args, format );
	wvsprintf( str, format, args );
	va_end( args );
	
	WriteFile( GetStdHandle( STD_OUTPUT_HANDLE ), str, lstrlen( str ), &ret, nullptr );
}


char* LastError( void )
{
	static char str[1024];
	FormatMessage( FORMAT_MESSAGE_FROM_SYSTEM | FORMAT_MESSAGE_IGNORE_INSERTS, nullptr, GetLastError(),
					MAKELANGID( LANG_NEUTRAL, SUBLANG_DEFAULT ), str, sizeof(str), nullptr );
	return GetLastError() ? str : (char*)"no error";
}







///////////////////////////////////////////////////////////////////////////////
bool ReadTrack( FDimg& fd, BYTE cyl, BYTE head )
{
	SECINF secinf[MAX_SECTOR];
	DWORD ltime = 0;		// インデックスホールから前セクタ先頭までの時間(us)
	bool err = false;		// エラー検出フラグ
	int secnum = 0;			// セクタ数
	
	FDTRK& track = fd.Track( cyl, head );
	
	// 初期値はアンフォーマット(エラートラックはアンフォーマットとして扱う)
	track.Unformat();
	
	// 対象トラックをスキャン
	// 単密度チェック
	if( SCANSD ){
		secnum = fdc.ScanTrack( head, secinf, FD_OPTION_FM );
	}
	
	// 倍密度チェック
	int secmfm = fdc.ScanTrack( head, &secinf[secnum], FD_OPTION_MFM );
	
	if( secnum < 0 || secmfm < 0 ){
		// トラックのスキャンに失敗した場合はアンフォーマット扱いとして処理を続ける
		WriteCon( "<<Scan track error>>\n" );
		return true;
	}
	
	// 処理するセクタ数
	secnum += secmfm;
	
	if( !secnum ){
		if( VERBOSE ){ WriteCon( " <No sector>\n" ); }
		return true;
	}
	
	if( VERBOSE ){
		WriteCon( "Sectors:%2u\n", secnum );
		WriteCon( "\t[Sec] [C] [H] [R] [N]  [Size] [Time]\n" );
	}
	
	// セクタデータを準備
	SECINF* si = secinf;
	for( int i=0; i < secnum; i++, si++ ){
		FDSEC& sec = track.AddSector( si->c, si->h, si->r, si->n, si->density == FD_OPTION_MFM ? 0 : 0x40 );
		sec.SetRelTime( si->reltime );
	}
	
	// セクタ毎に読込み
	for( int i=1; i < secnum+1; i++ ){
		FDSEC& sec = track.Sector( i );
		if( VERBOSE ){ WriteCon( "\t  %2u: %3u %3u %3u %3u :%5u  %5u %s %s", i, sec.C(), sec.H(), sec.R(), sec.N(), sec.Size(), sec.RelTime()-ltime, sec.Density() ? "FM " : "MFM", sec.SectorOffset() ? "[DUP]" : "" ); }
		ltime = sec.RelTime();
		
		if( sec.N() > 8 ){
			WriteCon( " Over size\n" );
			break;
		}
		
		// セクタオフセット設定
		if( sec.SectorOffset() ){
			fdc.SetSectorOffset( sec.SectorOffset() - 1 );
		}
		
		if( fdc.CmdReadData( head, sec.C(), sec.H(), sec.R(), sec.N(), sec.Buffer(), sec.Density() ? FD_OPTION_FM : FD_OPTION_MFM ) ){
			// 正常終了
			sec.SetDeleted( fdc.ST2() & ST2_CONTROL_MARK ? 0x10 : 0 );
			sec.SetStatus( BIOS_READY );
			
			if( VERBOSE ){
				if( fdc.ST2() & ST2_CONTROL_MARK ){ WriteCon( "[Deleted]" ); }
				WriteCon( "-> OK\n" );
			}
		}else{
			// エラー終了
			// CRCエラー?
			if( fdc.ST1() & ST1_DATA_ERROR ){
				if( fdc.ST2() & ST2_DE_IN_DATA_FIELD ){		// DATA
					sec.SetStatus( BIOS_DATA_CRC_ERROR );
					if( VERBOSE ){ WriteCon( "-> DATA CRC Error\n" ); }
					else		 { WriteCon( "[DATA CRC Error]" ); }
				}else{										// ID
					sec.SetStatus( BIOS_ID_CRC_ERROR );
					if( VERBOSE ){ WriteCon( "-> ID CRC Error\n" ); }
					else		 { WriteCon( "[ID CRC Error]" ); }
				}
			// Missing DAM/DDAM?
			}else if( (fdc.ST1() & ST1_MISSING_AM) && (fdc.ST2() & ST2_MA_IN_DATA_FIELD) ){
				sec.SetStatus( BIOS_MISSING_DAM );
				if( VERBOSE ){ WriteCon( "-> Missing DAM\n" ); }
				else		 { WriteCon( "[Missing DAM]" ); }
			// それ以外のエラーの場合はとりあえず0xffにしておく
			}else{
				sec.SetStatus( 0xff );
				if( VERBOSE ){ WriteCon( "-> Error\n" ); }
				else		 { WriteCon( "[Error]" ); }
			}
			err = true;
		}
	}
	if( !VERBOSE && err ){ WriteCon( "\n" ); }
	
	return true;
}


///////////////////////////////////////////////////////////////////////////////
BYTE Byte2N( DWORD size )
{
	BYTE sz = 0;
	while( (size>>sz) > 128 ) sz++;
	return sz;
}


///////////////////////////////////////////////////////////////////////////////
bool WriteTrack( FDimg& fd, BYTE cyl, BYTE head )
{
	FDTRK& track = fd.Track( cyl, head );
	
	if( !track.Sectors() ){
		if( VERBOSE ){ WriteCon( " <No sector>\n" ); }
		return true;
	}
	
	if( VERBOSE ){
		WriteCon( "Sectors:%2u%\n", track.Sectors() );
		WriteCon( "\t[Sec] [C] [H] [R] [N]  [Size]\n" );
	}
	
	// セクタ毎に書込み
	for( int i=1; i < track.Sectors()+1; i++ ){
		FDSEC& sec = track.Sector( i );
		if( VERBOSE ){ WriteCon( "\t  %2u: %3u %3u %3u %3u :%5u(%d) %s %s", i, sec.C(), sec.H(), sec.R(), sec.N(), sec.Size(), track.ActualSectorSize( i ), sec.Density() ? "FM " : "MFM", sec.SectorOffset() ? "[DUP]" : "" ); }
		
		// データ書き込み不要?
		// = Missing DAMセクタの場合
		// = フォーマットしたまま かつ DATA CRC ERRORセクタで有効セクタ長とNが一致しない
		bool unq =	( sec.Status() == BIOS_MISSING_DAM ) ||
					( sec.IsUniqueData() && track.EstimateGap3Data() == sec.Buffer()[0] &&
					  sec.Status() == BIOS_DATA_CRC_ERROR && track.ActualSectorSize( i ) != sec.N() );
		
		bool ret;
		if( unq ){
			if( sec.Status() == BIOS_DATA_CRC_ERROR && VERBOSE ){
				WriteCon( "[CRC error]" );
			}else if( sec.Status() == BIOS_MISSING_DAM && VERBOSE ){
				WriteCon( "[Missing DAM]" );
			}
			ret = true;
		}else{
			// DATA CRCエラー処理
			if( sec.Status() == BIOS_DATA_CRC_ERROR ){
				if( VERBOSE ){ WriteCon( "[CRC error]" ); }
				fdc.SetShortWrite( (128<<track.ActualSectorSize( i )) );
			}
			
			// セクタオフセット設定
			if( sec.SectorOffset() )
				fdc.SetSectorOffset( sec.SectorOffset() - 1 );
			
			if( sec.Deleted() ){
				if( VERBOSE ){ WriteCon( "[Deleted]" ); }
				ret = fdc.CmdWriteDeletedData( head, sec.C(), sec.H(), sec.R(), sec.N(), sec.Buffer(), sec.Density() ? FD_OPTION_FM : FD_OPTION_MFM );
			}else{
				ret = fdc.CmdWriteData( head, sec.C(), sec.H(), sec.R(), sec.N(), sec.Buffer(), sec.Density() ? FD_OPTION_FM : FD_OPTION_MFM );
			}
		}
		
		if( ret ){
			if( VERBOSE ){ WriteCon( "-> OK\n" ); }
		}else{
			WriteCon( "-> %s", LastError() );
		}
	}
	
	return true;
}


///////////////////////////////////////////////////////////////////////////////
bool UnformatTrack( BYTE head, BYTE flag=FD_OPTION_MFM )
{
	BYTE buf[sizeof(FD_FORMAT_PARAMS)+sizeof(FD_ID_HEADER)];
	
	PFD_FORMAT_PARAMS pfp = (PFD_FORMAT_PARAMS)buf;
	PFD_ID_HEADER ph      = pfp->Header;
	
	pfp->flags   = flag;
	pfp->phead   = head;
	pfp->size    = DiskType&FD_HIGHDENSITY ? 7 : 6;
	pfp->sectors = 1;
	pfp->gap     = 0;
	pfp->fill    = 0;
	
	ph->cyl      = 0;
	ph->head     = 0;
	ph->sector   = 0;
	ph->size     = 0;
	
	if( fdc.CmdFormat( pfp ) ){
		if( VERBOSE ){ WriteCon( " -> OK\n" ); }
	}else{
		WriteCon( " -> %s", LastError() );
	}
	
	return true;
}


///////////////////////////////////////////////////////////////////////////////
// プリアンブル直後から指定バイト消去する
// トラック先頭にCRCエラーセクタができるので再フォーマット(FDCリセットあり)が前提
///////////////////////////////////////////////////////////////////////////////
bool WipeArea( BYTE head, DWORD len, BYTE flag )
{
	if( VERBOSE ){ WriteCon( "<Wipe %ld>", len ); }
	fdc.SetShortWrite( 4, (len-12-4-4)*32 );
	
	return UnformatTrack( head, flag );
}


///////////////////////////////////////////////////////////////////////////////
bool FormatTrack( FDimg& fd, BYTE cyl, BYTE head )
{
	BYTE buf[sizeof(FD_FORMAT_PARAMS)+sizeof(FD_ID_HEADER)*MAX_SECTOR];
	PFD_FORMAT_PARAMS pfp = (PFD_FORMAT_PARAMS)buf;
	PFD_ID_HEADER ph      = pfp->Header;
	int nn = 1;
	
	FDTRK& track = fd.Track( cyl, head );
	
	if( !track.Sectors() ){
		return UnformatTrack( head );
	}
	
	// セクタ長混在チェック
	BYTE secmax = track.ActualSectorSize( 1 );
	BYTE secmin = secmax;
	for( int i=1; i <= track.Sectors(); i++ ){
		BYTE secact = track.ActualSectorSize( i );
		secmax = max( secmax, secact );
		secmin = min( secmin, secact );
	}
	
	// 記録密度は最初のセクタで判断
	pfp->flags   = track.Sector( 1 ).Density() == 0x40 ? FD_OPTION_FM : FD_OPTION_MFM;
	pfp->phead   = head;
	pfp->size    = secmin;
	pfp->sectors = 0;
	pfp->gap     = track.Gap3();
	pfp->fill    = track.EstimateGap3Data();
	
	for( int i=1; i <= track.Sectors(); i++ ){
		FDSEC& sec = track.Sector( i );
		// セクタ長混在してたら最短長セクタに分割(Nではなく有効データサイズで判断)
		nn = track.SpritSector( i, secmin );
		for( int j=0; j < nn; j++, ph++ ){
			ph->cyl    = j ? 0      : sec.C();
			ph->head   = j ? 0      : sec.H();
			ph->sector = j ? 0      : sec.R();
			ph->size   = j ? secmin : sec.N();
			pfp->sectors++;
		}
	}
	
	if( VERBOSE ){
		if( secmax == secmin ){
			WriteCon( "Sectors:%2u Size:%u", pfp->sectors, pfp->size );
		}else{
			WriteCon( "Sectors:%2u(%2u) Size:%u-%u[Mixed]", track.Sectors(), pfp->sectors, pfp->size, secmax );
		}
		WriteCon( " %s Gap:%02XH Data:%02XH\n", pfp->flags == FD_OPTION_MFM ? "MFM" : "FM ", pfp->gap, pfp->fill );
		
		WriteCon( "\t[Sec] [C] [H] [R] [N]  [Size]\n" );
		PFD_ID_HEADER phh = pfp->Header;
		for( int i=1; i <= track.Sectors(); i++ ){
			nn = track.SpritSector( i, secmin );
			for( int j=0; j < nn; j++, phh++ ){
				if( j ){ WriteCon( "\t      %3u %3u %3u %3u :%5u Dummy\n", phh->cyl, phh->head, phh->sector, phh->size, 128<<pfp->size ); }
				else   { WriteCon( "\t  %2u: %3u %3u %3u %3u :%5u\n", i, phh->cyl, phh->head, phh->sector, phh->size, 128<<pfp->size ); }
			}
		}
	}
	
	// エラーセクタ処理 その1
	// 最終セクタのステータスチェック
	// Missing DAM または ID CRC ERROR の場合はリセット指示
	FDSEC& sec = track.Sector( track.Sectors() );
	if( sec.Status() == BIOS_ID_CRC_ERROR ){
		fdc.SetShortWrite( 4 * (pfp->sectors - nn + 1) );			// ID
	}else if( sec.Status() == BIOS_MISSING_DAM ){
		fdc.SetShortWrite( 4 * (pfp->sectors - nn + 1), 32 * 3 );	// ID+CRC+GAP2
	}
	
	// フォーマット実行
	if( !fdc.CmdFormat( pfp ) ){ return false; }
	
	
	// エラーセクタ処理 その2
	// ID CRC ERROR と Missing DAM
	int lastsec   = track.Sectors();
	pfp->sectors -= nn;
	while( --lastsec > 0 ){
		FDSEC& sec = track.Sector( lastsec );
		nn = track.SpritSector( lastsec, secmin );
		
		if( (sec.Status() == BIOS_ID_CRC_ERROR) || (sec.Status() == BIOS_MISSING_DAM) ){
			DWORD misc = pfp->flags == FD_OPTION_FM ? 6+1+4+2+11+6+1+2 : 12+4+4+2+22+12+4+2;
			DWORD dlen = ( (128<<secmin) + misc + pfp->gap ) * pfp->sectors - pfp->gap;
			
			// 一旦ここまで消去
			if( !WipeArea( head, dlen, pfp->flags ) ){ return false; }
			
			pfp->sectors -= nn - 1;
			if( sec.Status() == BIOS_ID_CRC_ERROR ){
				fdc.SetShortWrite( 4 * pfp->sectors );			// ID
			}else if( sec.Status() == BIOS_MISSING_DAM ){
				fdc.SetShortWrite( 4 * pfp->sectors, 32 * 3 );	// ID+CRC+GAP2
			}
			
			// 改めてフォーマット
			if( !fdc.CmdFormat( pfp ) ){ return false; }
		}
		pfp->sectors--;
	}
	
	return true;
}


///////////////////////////////////////////////////////////////////////////////
bool ReadDisk( FDimg& fd, BYTE strk, BYTE etrk )
{
	int side = fd.Type() & FD_DOUBLESIDE  ? 2 : 1;
	int step = fd.Type() & FD_DOUBLETRACK ? 1 : 2;
	
	for( int cyl = strk; cyl <= etrk; cyl += step ){
		if( !fdc.CmdSeek( cyl ) ){ return false; }
		
		for( int head = 0; head < side; head++ ){
			WriteCon( "\rRead     Cylinder:%2u Track %2u Side %u ", cyl, cyl/step, head );
			if( !ReadTrack( fd, cyl, head ) ){
				// エラーの場合はメッセージを表示するが処理は継続する
				WriteCon( "\n!!Read Error: %s\n", LastError() );
			}
		}
	}
	return true;
}


///////////////////////////////////////////////////////////////////////////////
bool WriteDisk( FDimg& fd, BYTE strk, BYTE etrk )
{
	int side  = DiskType  & FD_DOUBLESIDE  ? 2 : 1;
	int stepd = DiskType  & FD_DOUBLETRACK ? 1 : 2;
	int steps = fd.Type() & FD_DOUBLETRACK ? 1 : 2;
	
	for( int cyl = strk; cyl <= etrk; cyl++ ){
		if( !fdc.CmdSeek( cyl ) ){ return false; }
		
		for( int head = 0; head < 2; head++ ){
			if( (head < side ) && !((cyl & 1)&&(stepd == 2)) ){
				// 必ずアンフォーマットする
				WriteCon( "\rUnformat Cylinder:%2u Track:%2u Side:%u ", cyl, cyl/stepd, head );
				if( !UnformatTrack( head ) ){
					WriteCon( "\n!!Unformat Error: %s\n", LastError() );
					continue;
				}
				
				BYTE tcyl = (cyl*steps)/stepd;
				if( fd.Track( tcyl, head ).Sectors() ){	// Write
					WriteCon( "\rFormat   Cylinder:%2u Track:%2u Side:%u ", cyl, cyl/stepd, head );
					if( !FormatTrack( fd, tcyl, head ) ){
						WriteCon( "\n!!Format Error: %s\n", LastError() );
						continue;
					}
					
					WriteCon( "\rWrite    Cylinder:%2u Track:%2u Side:%u ", cyl, cyl/stepd, head );
					if( !WriteTrack ( fd, tcyl, head ) ){
						WriteCon( "\n!!rWrite Error: %s\n", LastError() );
						continue;
					}
				}
			}else if( UNNEEDED ){
				WriteCon( "\rUnformat Cylinder:%2u Track:%2u Side:%u ", cyl, cyl/stepd, head );
				if( !UnformatTrack( head ) ){ return false; }
			}
		}
	}
	return true;
}


///////////////////////////////////////////////////////////////////////////////
bool UnformatDisk( BYTE fdtype, BYTE strk, BYTE etrk )
{
	int side = fdtype&FD_DOUBLESIDE  ? 2 : 1;
	int step = fdtype&FD_DOUBLETRACK ? 1 : 2;
	
	for( int cyl = strk; cyl <= etrk; cyl++ ){
		for( int head = 0; head < 2; head++ ){
			if( UNNEEDED ? ((head == 1)&&(side == 1)) || ((cyl & 1)&&(step == 2))
						 : (head < side ) && !((cyl & 1)&&(step == 2)) ){
				if( !fdc.CmdSeek( cyl ) ){ return false; }
				WriteCon( "\rUnformat Cylinder:%2u Track:%2u Side:%u ", cyl, cyl/step, head );
				if( !UnformatTrack( head ) ){ return false; }
			}
		}
	}
	return true;
}







///////////////////////////////////////////////////////////////////////////////
bool IsFDD( const std::string& str, int& pDrvNum )
{
	if( !isalpha( str.c_str()[0] ) || str.c_str()[1] != ':' || str.c_str()[2] ){
		return false;
	}
	
	pDrvNum = (str.c_str()[0] & ~0x20) - 'A';
	
	return true;
}


///////////////////////////////////////////////////////////////////////////////
bool CheckArg( int argc, char** argv, CmdType& pcom, int& pdrv, BYTE& pfdt,
				std::string& title, int& retry, BYTE& opt, std::string& fname )
{
	if( argc < 3 ){ return false; }
	
	pcom  = cmdNone;
	pfdt  = FDUNKNOWN;
	opt   = 0;
	
	// 第1引数 (コマンド)
	if     ( !lstrcmpi( argv[1], "r" ) ){ pcom = cmdRead; }
	else if( !lstrcmpi( argv[1], "w" ) ){ pcom = cmdWrite; }
	else if( !lstrcmpi( argv[1], "u" ) ){ pcom = cmdUnformat; }
	else								{ return false; }
	
	// 第2引数 (ディスクの種類)
	if     ( !lstrcmpi( argv[2], "1D"  ) ){ pfdt = FD1D; }
	else if( !lstrcmpi( argv[2], "2D"  ) ){ pfdt = FD2D; }
	else if( !lstrcmpi( argv[2], "1DD" ) ){ pfdt = FD1DD; }
	else if( !lstrcmpi( argv[2], "2DD" ) ){ pfdt = FD2DD; }
	else if( !lstrcmpi( argv[2], "2HD" ) ){ pfdt = FD2HD; }
	else								  { return false; }
	
	// 第3引数以降は順不同
	for( int i=3; i < argc; i++ ){
		if     ( IsFDD( argv[i], pdrv ) && pdrv > 1 ){ return false; }
		else if( argv[i][0] == '-'){
			switch( std::toupper(argv[i][1]) ){
			case 'T': title = &argv[i][2];         break;
			case 'R': retry = std::atoi( &argv[i][2] ); break;
			case 'C':
				switch( std::toupper(argv[i][2]) ){
				case 'S': SCyl = std::atoi( &argv[i][3] ); break;
				case 'E': ECyl = std::atoi( &argv[i][3] ); break;
				default:  return false;
				}
				break;
			default:
				for( int j=1; j < lstrlen( argv[i] ); j++ ){
					switch( std::toupper(argv[i][j]) ){
					case 'W': opt |= OPT_PROTECT;   break;
					case 'O': opt |= OPT_OVERTRACK; break;
					case '5': opt |= OPT_D5INCH;    break;
					case 'S': opt |= OPT_SCANSD;    break;
					case 'D': opt |= OPT_1DDITT;    break;
					case 'U': opt |= OPT_UNNEEDED;  break;
					case 'V': opt |= OPT_VERBOSE;   break;
					default:  return false;
					}
				}
			}
		}else{
			fname = argv[i];
			// 拡張子が指定されていなければデフォルト拡張子を設定
			if( *PathFindExtension( fname.c_str() ) != '.' ){
				PathRenameExtension( (char*)fname.c_str(), DEFAULTEXT );
			}
		}
	}
	
	if( ECyl == 255 ){
		ECyl = pfdt & FD_DOUBLETRACK ? 79 : 39;
	}
	
	if( !(pfdt & FD_DOUBLETRACK) ){
		SCyl *= 2;
		ECyl  = ECyl*2+1;	// 奇数トラックまで処理する
	}
	
	ECyl = OVERTRACK ? MAX_CYLINDER-1 : ECyl;
	
	if( (pcom == cmdNone) || (pfdt == FDUNKNOWN) || (SCyl > ECyl) || (SCyl < 0 ) || (ECyl >= MAX_CYLINDER ) ){
		return false;
	}
	
	return true;
}


///////////////////////////////////////////////////////////////////////////////
void Usage( void )
{
	WriteCon(	"NDitt Ver.0.05 -- Next-gen Ditt --\n"
				"\n"
				"Usage: NDitt <command> <type> [-option] <drive> <image>\n"
				"\n"
				"  command:  r - Create image from floppy disk\n"
				"            w - Write image to floppy disk\n"
				"            u - Unformat\n"
				"  type   :  Media type (1D,2D,1DD,2DD,2HD)\n"
				"  option :  -txxx - Disk title, 16 characters\n"
				"            -rxxx - Retry count, default:10\n"
				"            -csxx - Start cylinder, default:0\n"
				"            -cexx - End   cylinder, default:39 or 79\n"
				"            -w    - Write protect\n"
				"            -o    - Over track, Track:160-163\n"
				"            -5    - 5inch drive\n"
				"            -s    - Scan single density track\n"
				"            -d    - 1DDITT format (1DD read)\n"
				"            -u    - Unformat unneeded track(Write/Unformat)\n"
				"            -v    - Verbose\n"
				"  drive  :  FD drive, typically A: or B:\n"
				"  image  :  Disk image to create or write (.d88)\n"
	);
}








///////////////////////////////////////////////////////////////////////////////
int main( int argc_, char* argv_[] )
{
	CmdType nCommand = cmdNone;
	int DrvNum = -1;
	std::string fname = "";
	std::string title = "";
	D88DISK fdi;
	int retry = 10;
	DatRate DataRate;
	int res;
	
	fname.reserve( MAX_PATH );
	
	
	// 引数解析
	if( !CheckArg( argc_, argv_, nCommand, DrvNum, DiskType, title, retry, Option, fname ) ){
		Usage();
		return EXIT_FAILURE;
	}
	
	// 初期設定(ディスクの名前,ディスクの種類,リトライ回数,転送レート)
	fdi.SetTitle( title );
	fdi.SetType( DiskType );
	fdc.SetRetry( retry );
	DataRate = DiskType & FD_HIGHDENSITY ? Rate500K : D5INCH ? Rate300K : Rate250K;
	
	// 初期化
	if( (res = fdc.Init( DrvNum, DataRate )) ){
		WriteCon( "\nFailed : " );
		switch( res ){
		case 1: WriteCon( "fdrawcmd.sys is not installed, see: http://simonowen.com/fdrawcmd/\n" );	break;
		case 2: WriteCon( "The installed fdrawcmd.sys is not compatible with this utility.\n" );	break;
		case 3: WriteCon( "Failed to open FD drive: %s", LastError() );								break;
		case 4: WriteCon( "Failed to initialise controller: %s", LastError() );						break;
		case 5: WriteCon( "No media in drive.\n" );													break;
		default:;
		}
		return EXIT_FAILURE;
	}
	
	WriteCon( "Command    : %s\n", comname[nCommand].c_str() );
	WriteCon( "Media type : %c%c%c%s\n", DiskType&FD_DOUBLESIDE  ? '2' : '1',
										 DiskType&FD_HIGHDENSITY ? 'H' : DiskType&FD_DOUBLEDENSITY ? 'D' : 'S',
										 DiskType&FD_DOUBLETRACK ? 'D' : ' ', D5INCH ? "(5inch)" : "" );
	WriteCon( "Drive      : %c:\n", 'A' + DrvNum );
	WriteCon( "Data rate  : %s\n", bpsname[DataRate].c_str() );
	
	
	// コマンド別処理
    switch( nCommand ){
	case cmdRead:		// Read -------------------------------------------------
		WriteCon( "File Name  : %s\n", fname.c_str() );
		WriteCon( "Title      : %s\n", title.c_str() );
		
		if( !ReadDisk( fdi, SCyl, ECyl ) ){
			WriteCon( "\n\nRead failed: %s\n", LastError() );
			return EXIT_FAILURE;
		}
		
		if( !fdi.WriteToFile( fname, PROTECT | F1DDITT ) ){
			WriteCon( "\n\nWrite image file failed: %s\n", LastError() );
			return EXIT_FAILURE;
		}
		
		WriteCon( "\nImage created successfully.\n" );
		break;
		
	case cmdWrite:		// Write ------------------------------------------------
		WriteCon( "File Name  : %s\n", fname.c_str() );
		
		if( !fdi.ReadFromFile( fname ) ){
			WriteCon( "\n\nRead image file failed: %s\n", LastError() );
			return EXIT_FAILURE;
		}
		
		if( !WriteDisk( fdi, SCyl, ECyl ) ){
			WriteCon( "\n\nWrite failed: %s\n", LastError() );
			return EXIT_FAILURE;
		}
		
		WriteCon( "\nImage written successfully.\n" );
		break;
		
	case cmdUnformat:	// Unformat ---------------------------------------------
		if( UnformatDisk( DiskType, SCyl, ECyl ) )
			WriteCon( "\rUnformat disk successfully.\n" );
		else{
			WriteCon( "\n\nUnformat failed: %s\n", LastError() );
			return EXIT_FAILURE;
		}
		break;
		
	default:
		break;
	}
	
	
	
	return EXIT_SUCCESS;
}
