﻿#include <algorithm>
#include <cstring>

#include "fdimg.h"
#include "d88.h"


#define	GAP4LENGTH		50		// GAP4長(MFM) 正しい数値はよくわからない
#define	SECTOR_FILL		0xff	// fill byte for formatted sectors



FDSEC DummySEC( 0, 0, 0, 0 );	// ダミーセクタ
FDTRK DummyTRK;					// ダミートラック


// -------------------------------------------------------------
// セクタクラス
// -------------------------------------------------------------
////////////////////////////////////////////////////////////////
// コンストラクタ
////////////////////////////////////////////////////////////////
FDSEC::FDSEC( BYTE cc, BYTE hh, BYTE rr, BYTE nn, BYTE den, BYTE del, BYTE sta ) : 
	c( cc ), h( hh ), r( rr ), n( nn ), density( den ), deleted( del ), status( sta ),
	secoff( 0 ), reltime( 0 )
{
	if( nn == 0xff ){ data.clear(); }
	else            { data.assign( 128 << (( nn > 8 ) ? 8 : nn), 0 ); }
}


////////////////////////////////////////////////////////////////
// デストラクタ
////////////////////////////////////////////////////////////////
FDSEC::~FDSEC( void )
{
}


////////////////////////////////////////////////////////////////
// 各種パラメータ取得・設定
////////////////////////////////////////////////////////////////
const BYTE FDSEC::C( void ) const{ return c; }				// C取得
const BYTE FDSEC::H( void ) const{ return h; }				// H取得
const BYTE FDSEC::R( void ) const{ return r; }				// R取得
const BYTE FDSEC::N( void ) const{ return n; }				// N取得
const BYTE FDSEC::Density( void ) const{ return density; }	// 記録密度取得
void FDSEC::SetDensity( BYTE de ){ density = de; }			// 記録密度設定
const BYTE FDSEC::Deleted( void ) const{ return deleted; }	// DELETED MARK取得
void FDSEC::SetDeleted( BYTE de ){ deleted = de; }			// DELETED MARK設定
const BYTE FDSEC::Status( void ) const{ return status; }	// ステータス取得
void FDSEC::SetStatus( BYTE st ){ status = st; }			// ステータス設定
const WORD FDSEC::Size( void ) const{ return data.size(); }	// データサイズ取得
BYTE* FDSEC::Buffer( void ){ return data.data(); }			// データへのポインタ取得
int FDSEC::SectorOffset( void ){ return secoff; }			// セクタオフセット取得
void FDSEC::SetSectorOffset( int so ){ secoff = so; }		// セクタオフセット設定
DWORD FDSEC::RelTime( void ){ return reltime; }				// インデックスホールからの時間(us)取得
void FDSEC::SetRelTime( DWORD lt ){ reltime = lt; }			// インデックスホールからの時間(us)設定


///////////////////////////////////////////////////////////////////////////////
// 単一データセクタ?
///////////////////////////////////////////////////////////////////////////////
bool FDSEC::IsUniqueData( void )
{
	if( data.empty() ) return false;
	
	bool ret = true;
	std::for_each( data.begin(), data.end(), [&]( BYTE& dat ){
		if( dat != data[0] ){ ret = false; }
	});
	
	return ret;
}


// -------------------------------------------------------------
// トラッククラス
// -------------------------------------------------------------
////////////////////////////////////////////////////////////////
// コンストラクタ
////////////////////////////////////////////////////////////////
FDTRK::FDTRK( void ) : fdtype( FDUNKNOWN )
{
}


////////////////////////////////////////////////////////////////
// デストラクタ
////////////////////////////////////////////////////////////////
FDTRK::~FDTRK( void )
{
}


////////////////////////////////////////////////////////////////
// 各種パラメータ取得・設定
////////////////////////////////////////////////////////////////
void FDTRK::SetType( BYTE tp )			// ディスクの種類設定
{
	fdtype = tp;
}

int FDTRK::Sectors( void )				// セクタ数取得
{
	return sectors.size();
}


////////////////////////////////////////////////////////////////
// セクタ分割数を求める(ミックスセクタフォーマット時)
////////////////////////////////////////////////////////////////
int FDTRK::SpritSector( int nsec, BYTE secmin )
{
	if( (nsec < 1) || (nsec > Sectors()) ) return 0;
	
	int misc  = Sector( nsec ).Density() == 0x40 ? 6+1+4+2+11+6+1+2 : 12+4+4+2+22+12+4+2;
	int ssize = (128<<secmin) + misc + Gap3();
	int asize = -((128<<ActualSectorSize( nsec )) + misc + 2 );
	int snum  = 0;
	while( asize < 0 ){
		asize += ssize;
		snum++;
	}
	return snum;
}


////////////////////////////////////////////////////////////////
// GAP3長取得
////////////////////////////////////////////////////////////////
BYTE FDTRK::Gap3( void )
{
	if( !Sectors() ) return 0;
	
	// フォーマット時のセクタサイズを求める
	BYTE secmin = ActualSectorSize( 1 );
	for( int i=1; i <= Sectors(); i++ ){
		secmin = min( secmin, ActualSectorSize( i ) );
	}
	
	BYTE mgap = 0;
	int max   = 0;
	bool fm   = Sector( 1 ).Density() == 0x40;
	int tcap  = (TRACKCAP( fdtype ) -80-12-4-50 - GAP4LENGTH) / (fm ? 2 : 1);
	int misc  = fm ? 6+1+4+2+11+6+1+2 : 12+4+4+2+22+12+4+2;
	
	// GAP3の最大値はPC-98 BIOSの標準値に合わせて制限しておく
//	const BYTE gap3[][9] = {
//		{ 0x33, 0x33, 0x50, 0x74, 0x74, 0x74, 0x74, 0x74, 0x74 },	// MFM
//		{ 0x1b, 0x2a, 0x3a, 0x3a, 0x3a, 0x3a, 0x3a, 0x3a, 0x3a } };	// FM
	for( int gap3=2; gap3 < (fm ? 0x3a : 0x74); gap3++ ){
		int sgap  = 19999;
		int ssize = (128<<secmin) + misc + gap3;
		int tsize = 0;
		for( int i=1; i <= Sectors(); i++ ){
			int asize  = -((128<<ActualSectorSize( i )) + misc + 2);
			
			while( asize < 0 ){
				asize += ssize;
				tsize += ssize;
			}
			if( sgap >= asize ) sgap = asize;
		}
		
		if( tsize < tcap && max < sgap ){
			max  = sgap;
			mgap = gap3;
		}
	}
	return mgap;
}


///////////////////////////////////////////////////////////////////////////////
// GAP3データを推定する
// 単一データで埋め尽くされているセクタを検索
// DATA CRC ERRORセクタの数値を優先する
///////////////////////////////////////////////////////////////////////////////
BYTE FDTRK::EstimateGap3Data( void )
{
	BYTE gap3  = SECTOR_FILL;
	bool crces = false;			// DATA CRC ERRORセクタ検出フラグ
	
	std::for_each( sectors.begin(), sectors.end(), [&]( FDSEC& sec ){
		if( sec.IsUniqueData() ){
			if( sec.Status() == BIOS_DATA_CRC_ERROR ){
				gap3  = sec.Buffer()[0];
				crces = true;
			}else if( !crces )
				gap3 = sec.Buffer()[0];
		}
	});
	
	return gap3;
}


////////////////////////////////////////////////////////////////
// セクタオフセット設定
////////////////////////////////////////////////////////////////
void FDTRK::SetSectorOffset( void )
{
	if( Sectors() < 2 ) return;
	
	auto se = sectors.begin()+1;
	std::for_each( se, sectors.end(), [&]( FDSEC& sec0 ){
		int so = 0;
		std::for_each( sectors.begin(), se, [&]( FDSEC& sec1 ){
			if( (sec0.C() == sec1.C()) && (sec0.H() == sec1.H()) && (sec0.R() == sec1.R()) && (sec0.N() == sec1.N()) ){
				sec1.SetSectorOffset( ++so );
			}
		});
		sec0.SetSectorOffset( so );
		se++;
	});
}


////////////////////////////////////////////////////////////////
// セクタ取得
////////////////////////////////////////////////////////////////
FDSEC& FDTRK::Sector( const int sc )
{
	try{
		return sectors.at( sc - 1 );
	}
	catch( std::out_of_range& ){}
	
	return DummySEC;
}


////////////////////////////////////////////////////////////////
// セクタ追加
////////////////////////////////////////////////////////////////
FDSEC& FDTRK::AddSector( BYTE cc, BYTE hh, BYTE rr, BYTE nn, BYTE den, BYTE del, BYTE sta )
{
	sectors.emplace_back( cc, hh, rr, nn, den, del, sta );
	SetSectorOffset();
	
	return sectors.back();
}


////////////////////////////////////////////////////////////////
// フォーマット
////////////////////////////////////////////////////////////////
bool FDTRK::Format( BYTE sc, BYTE cc, BYTE hh, BYTE nn, BYTE den )
{
	Unformat();
	
	for( int i=1; i < sc+1; i++ ){
		AddSector( cc, hh, i, nn, den );
	}
	return true;
}


////////////////////////////////////////////////////////////////
// アンフォーマット
////////////////////////////////////////////////////////////////
void FDTRK::Unformat( void )
{
	sectors.clear();
}


////////////////////////////////////////////////////////////////
// 実際に有効と思われるセクタサイズ取得(DATA CRC ERRORセクタの場合)
////////////////////////////////////////////////////////////////
BYTE FDTRK::ActualSectorSize( int nsec )
{
	if( (nsec < 1) || (nsec > Sectors()) ) return 0xff;
	
	FDSEC& sec = Sector( nsec );
	FDSEC& nec = Sector( nsec + 1 );
	
	// "DATA CRC ERROR" かつ "最終セクタでない" 以外はNをそのまま返す
	if( !((sec.Status() == BIOS_DATA_CRC_ERROR) && nec.Density() != FD_UNKNOWN) )
		return sec.N();
	
	// 検索パターン作成(次のセクタのID周辺)
	BYTE strs[] = { 0xff, 0,0,0,0,0,0,             0xfe,                   0,0,0,0 };
	BYTE strd[] = { 0x4e, 0,0,0,0,0,0,0,0,0,0,0,0, 0xa1, 0xa1, 0xa1, 0xfe, 0,0,0,0 };
	bool fm = sec.Density() == 0x40;
	BYTE* pst = fm ? &strs[8] : &strd[17];
	*(pst++) = nec.C();
	*(pst++) = nec.H();
	*(pst++) = nec.R();
	*(pst++) = nec.N();
	pst = fm ? strs : strd;
	
	BYTE* sdata = sec.Buffer();
	BYTE sz     = 0;
	DWORD dlen  = 0;
	
	// reltimeがあれば時間で判定(実DISKから読み取ったデータの場合)
	if( sec.RelTime() && nec.RelTime() ){
		DWORD stime = nec.RelTime() - sec.RelTime();
		DWORD rate  = fdtype & FD_HIGHDENSITY ? 500 : 250;
		DWORD misc  = fm ? 6+1+4+2+11+6+1+2 : 12+4+4+2+22+12+4+2;
		
		dlen  = stime * (fm ? rate/2 : rate) / 8000 - misc - 1;	// DATA+GAP3
		
		// GAP3とDATA CRCを除く
		while( sdata[dlen] == pst[0] ) dlen--;
		dlen--;
		// サイズをNに変換
		while( dlen/(128<<sz) > 1 ) sz++;
		
	// reltimeが無ければデータ比較(イメージファイルの場合)
	}else{
		DWORD ssize = fm ? sizeof(strs) : sizeof(strd);
		DWORD dsize = (128<<((sec.N()>8) ? 8 : sec.N())) - ssize;
		
		do{
			if( std::memcmp( &sdata[dlen], pst, ssize ) == 0 ) break;
			dlen++;
		}while( dlen < dsize );
		
		if( dlen < dsize ){
			// GAP3とDATA CRCを除く
			while( sdata[dlen] == pst[0] ) dlen--;
			dlen--;
			// サイズをNに変換
			while( dlen/(128<<sz) > 1 ) sz++;
		}else{
			// 検索パターンが見つからなければ全データ有効と判断する
			sz = sec.N();
		}
	}
	return sz;
}


// -------------------------------------------------------------
// フロッピーディスクイメージクラス
// -------------------------------------------------------------
////////////////////////////////////////////////////////////////
// コンストラクタ
////////////////////////////////////////////////////////////////
FDimg::FDimg( void ) : title( "" ), fdtype( FDUNKNOWN ), protect( false )
{
	track.assign( MAX_CYLINDER*2, FDTRK() );
}


////////////////////////////////////////////////////////////////
// デストラクタ
////////////////////////////////////////////////////////////////
FDimg::~FDimg( void )
{
}


////////////////////////////////////////////////////////////////
// 各種パラメータ取得・設定
////////////////////////////////////////////////////////////////
const std::string& FDimg::Title( void ) const		// ディスクの名前取得
{
	return title;
}

void FDimg::SetTitle( const std::string& t )		// ディスクの名前設定
{
	title = t;
}

const BYTE FDimg::Type( void ) const				// ディスクの種類取得
{
	return fdtype;
}

void FDimg::SetType( BYTE tp )						// ディスクの種類設定
{
	fdtype = tp;
	std::for_each( track.begin(), track.end(), [&]( FDTRK& trk ){
		trk.SetType( fdtype );
	});
}

const bool FDimg::Protect( void ) const				// ライトプロテクト取得
{
	return protect;
}

void FDimg::SetProtect( bool wp )					// ライトプロテクト設定
{
	protect = wp;
}


////////////////////////////////////////////////////////////////
// 片面ディスク?
////////////////////////////////////////////////////////////////
bool FDimg::IsSingleSide( void )
{
	for( size_t i=1; i < track.size(); i += 2 )
		if( track[i].Sectors() ) return false;
	return true;
}


////////////////////////////////////////////////////////////////
// トラック取得
////////////////////////////////////////////////////////////////
FDTRK& FDimg::Track( BYTE cyl, BYTE head )
{
	try{
		return track.at( cyl * 2 + (head & 1) );
	}
	catch( std::out_of_range& ){}
	
	return DummyTRK;
}


////////////////////////////////////////////////////////////////
// アンフォーマット
////////////////////////////////////////////////////////////////
void FDimg::Unformat( void )
{
	std::for_each( track.begin(), track.end(), [&]( FDTRK& trk ){
		trk.Unformat();
	});
}


// -------------------------------------------------------------
// D88ディスクイメージクラス
// -------------------------------------------------------------
////////////////////////////////////////////////////////////////
// コンストラクタ
////////////////////////////////////////////////////////////////
D88DISK::D88DISK( void )
{
}


////////////////////////////////////////////////////////////////
// デストラクタ
////////////////////////////////////////////////////////////////
D88DISK::~D88DISK( void )
{
}


////////////////////////////////////////////////////////////////
// トラックデータ読込み
////////////////////////////////////////////////////////////////
bool D88DISK::ReadFileTrack( std::fstream& fs, int trk, FDTRK& track )
{
	D88SECTINFO sinfo;
	DWORD offset;
	int i=0;
	
	if( trk >= MAX_CYLINDER*2 ){ return false; }
	
	track.Unformat();
	
	// トラックテーブル読込み
	fs.seekg( 32 + trk * sizeof(DWORD), std::ios_base::beg );
	fs.read( (char*)&offset, sizeof(DWORD) );
	if( offset ){
		fs.seekg( offset, std::ios_base::beg );
		do{
			// セクタ情報読込み
			fs.read( (char*)&sinfo, sizeof(D88SECTINFO) );
			FDSEC& sec = track.AddSector( sinfo.c, sinfo.h, sinfo.r, sinfo.n, sinfo.density, sinfo.deleted, sinfo.status );
			
			// セクタデータ読込み
			fs.read( (char*)sec.Buffer(), sec.Size() );
		}while( ++i < sinfo.sec_nr );
	}
	return true;
}


////////////////////////////////////////////////////////////////
// トラックデータ書込み
////////////////////////////////////////////////////////////////
bool D88DISK::WriteFileTrack( std::fstream& fs, int trk, FDTRK& track )
{
	D88SECTINFO sinfo;
	DWORD offset=0;
	
	if( trk >= MAX_CYLINDER*2 ) return false;
	
	if( track.Sectors() ){
		fs.flush();	// 書込み時はseek前にflushしておくのがいいみたい
		fs.seekp( 0, std::ios_base::end );
		offset = fs.tellp();
		
		for( int i=1; i < track.Sectors()+1; i++ ){
			FDSEC& sec = track.Sector( i );
			// セクタ情報書込み
			ZeroMemory( &sinfo, sizeof(D88SECTINFO) );
			sinfo.c       = sec.C();			// ID の C (シリンダNo 片面の場合は=トラックNo)
			sinfo.h       = sec.H();			// ID の H (ヘッダアドレス 片面の場合は=0)
			sinfo.r       = sec.R();			// ID の R (トラック内のセクタNo)
			sinfo.n       = sec.N();			// ID の N (セクタサイズ 0:128 1:256 2:512 3:1024)
			sinfo.sec_nr  = track.Sectors();	// このトラック内に存在するセクタの数
			sinfo.density = sec.Density();		// 記録密度     0x00:倍密度   0x40:単密度
			sinfo.deleted = sec.Deleted();		// DELETED MARK 0x00:ノーマル 0x10:DELETED
			sinfo.status  = sec.Status();		// ステータス
			sinfo.size	  = sec.Size();			// このセクタ部のデータサイズ
			fs.write( (char*)&sinfo, sizeof(D88SECTINFO) );
			
			// セクタデータ書込み
			fs.write( (char*)sec.Buffer(), sinfo.size );
		}
	}
	
	// トラックテーブル書込み
	fs.flush();	// 書込み時はseek前にflushしておくのがいいみたい
	fs.seekp( 32 + trk * sizeof(DWORD), std::ios_base::beg );
	fs.write( (char*)&offset, sizeof(DWORD) );
	
	return true;
}


////////////////////////////////////////////////////////////////
// イメージファイルから読込み
//  1D=>2D,1DD=>2DDとして読込まれるので注意(可能ならば補正してみる)
////////////////////////////////////////////////////////////////
bool D88DISK::ReadFromFile( const std::string& fname )
{
	D88HEADER d88h;
	std::fstream fs;
	
	fs.open( fname, std::ios_base::in|std::ios_base::binary );
	if( !(fs.is_open() && fs.good()) ){ return false; }
	
	// ヘッダ情報読込み
	fs.read( (char*)&d88h, sizeof(D88HEADER) );
	d88h.name[16] = 0;
	title = (const char*)d88h.name;
	switch( d88h.type ){
	case 0x00:	fdtype = FD2D;		break;
	case 0x10:	fdtype = FD2DD;		break;
	case 0x20:	fdtype = FD2HD;		break;
	case 0x40:	fdtype = FD1DD;		break;	// 1DDitt
	default:	fdtype = FDUNKNOWN;
	}
	protect = d88h.protect;			// ライトプロテクト
	
	// トラック読込み
	if( fdtype == FD1DD ){	// 1DDittイメージは別処理
		for( int cyl = 0; cyl < MAX_CYLINDER; cyl++ ){
			ReadFileTrack( fs, cyl, track[cyl*2] );
		}
	}else{
		int step = fdtype & FD_DOUBLETRACK ? 1 : 2;
		for( int cyl = 0; cyl < MAX_CYLINDER; cyl += step ){
			for( int head = 0; head < 2; head++ ){
				ReadFileTrack( fs, cyl*2/step+head, track[cyl*2+head] );
			}
		}
	}
	
	fs.close();
	
	// 片面ディスクの場合
	if( !(fdtype&FD_HIGHDENSITY) && IsSingleSide() )
		fdtype &= ~FD_DOUBLESIDE;
	
	return true;
}


////////////////////////////////////////////////////////////////
// イメージファイルに書込み
////////////////////////////////////////////////////////////////
bool D88DISK::WriteToFile( const std::string& fname, BYTE flag )
{
	D88HEADER d88h;
	std::fstream fs;
	
	protect = (flag & FD_PROTECT) ? true : false;
	
	fs.open( fname, std::ios_base::out|std::ios_base::binary );
	if( !(fs.is_open() && fs.good()) ){ return false; }
	
	// ヘッダ情報書込み
	ZeroMemory( &d88h, sizeof(D88HEADER) );
	std::strncpy( (char*)d88h.name, title.c_str(), 16 );
	d88h.protect = protect ? 0x10 : 0;
	switch( fdtype ){
	case FD1D:	d88h.type = 0x00;	break;
	case FD2D:	d88h.type = 0x00;	break;
	case FD1DD:	d88h.type = (flag & FD_1DDITT) ? 0x40 : 0x10;	break;
	case FD2DD:	d88h.type = 0x10;	break;
	case FD2HD:	d88h.type = 0x20;	break;
	default:	d88h.type = 0x00;
	}
	fs.write( (char*)&d88h, sizeof(D88HEADER) );
	
	// トラック書込み
	if( d88h.type == 0x40 ){	// 1DDittイメージは別処理
		for( int cyl = 0; cyl < MAX_CYLINDER; cyl++ ){
			WriteFileTrack( fs, cyl, track[cyl*2] );
		}
	}else{
		int step = fdtype & FD_DOUBLETRACK ? 1 : 2;
		for( int cyl = 0; cyl < MAX_CYLINDER; cyl += step ){
			for( int head = 0; head < 2; head++ ){
				WriteFileTrack( fs, cyl*2/step+head, track[cyl*2+head] );
			}
		}
	}
	
	// ファイルサイズ書込み
	fs.flush();
	fs.seekp( 0, std::ios_base::end );
	d88h.size = fs.tellp();
	fs.seekp( 0, std::ios_base::beg );
	fs.write( (char*)&d88h, 32 );
	
	fs.close();
	
	return true;
}

