﻿#ifndef FDIMG_H_INCLUDED
#define FDIMG_H_INCLUDED

#include <fstream>
#include <string>
#include <vector>

#include "tdef.h"


#define FD_DOUBLESIDE			0x01	// 両面
#define	FD_DOUBLETRACK			0x02	// 倍トラック
#define	FD_DOUBLEDENSITY		0x04	// 倍密度
#define	FD_HIGHDENSITY			0x08	// 高密度
#define FD_UNKNOWN				0x00

#define	FD1D					(FD_DOUBLEDENSITY)
#define	FD2D					(FD_DOUBLESIDE  | FD_DOUBLEDENSITY)
#define	FD1DD					(FD_DOUBLETRACK | FD_DOUBLEDENSITY)
#define	FD2DD					(FD_DOUBLESIDE  | FD_DOUBLETRACK | FD_DOUBLEDENSITY)
#define	FD2HD					(FD_DOUBLESIDE  | FD_DOUBLETRACK | FD_HIGHDENSITY)
#define	FDUNKNOWN				0x00

#define	TRACKCAP(t)				((t) & FD_HIGHDENSITY ? 12500 : (t) & FD_DOUBLEDENSITY ? 6250 : 3125)

#define	FD_PROTECT				0x01
#define	FD_1DDITT				0x20

// Disk BIOS Status
#define BIOS_READY				(0x00)
#define BIOS_WRITE_PROTECT		(0x70)
#define BIOS_ID_CRC_ERROR		(0xa0)
#define BIOS_DATA_CRC_ERROR		(0xb0)
#define BIOS_NO_DATA			(0xc0)
#define BIOS_MISSING_IAM		(0xe0)
#define BIOS_MISSING_DAM		(0xf0)



#define	MAX_CYLINDER			82


////////////////////////////////////////////////////////////////
// クラス定義
////////////////////////////////////////////////////////////////
// セクタクラス
class FDSEC {
private:
	BYTE c;							// ID の C (シリンダNo 片面の場合は=トラックNo)
	BYTE h;							// ID の H (ヘッダアドレス 片面の場合は=0)
	BYTE r;							// ID の R (トラック内のセクタNo)
	BYTE n;							// ID の N (セクタサイズ 0:128 1:256 2:512 3:1024)
	BYTE density;					// 記録密度     0x00:倍密度   0x40:単密度
	BYTE deleted;					// DELETED MARK 0x00:ノーマル 0x10:DELETED
	BYTE status;					// ステータス
	
	std::vector<BYTE> data;			// データ
	
	int secoff;						// セクタオフセット(同一トラック内に同IDのセクタがある場合の通し番号)
									// 1個しかなければ0, 2個以上ある場合は1,2,3...
	DWORD reltime;					// インデックスホールからの時間(us)

public:
	FDSEC( BYTE, BYTE, BYTE, BYTE, BYTE = FD_UNKNOWN, BYTE = 0, BYTE = BIOS_READY );	// コンストラクタ
	~FDSEC();						// デストラクタ
	
	const BYTE C() const;			// C取得
	const BYTE H() const;			// H取得
	const BYTE R() const;			// R取得
	const BYTE N() const;			// N取得
	const BYTE Density() const;		// 記録密度取得
	void SetDensity( BYTE );		// 記録密度設定
	const BYTE Deleted() const;		// DELETED MARK取得
	void SetDeleted( BYTE );		// DELETED MARK設定
	const BYTE Status() const;		// ステータス取得
	void SetStatus( BYTE );			// ステータス設定
	const WORD Size() const;		// データサイズ取得
	BYTE* Buffer();					// データへのポインタ取得
	
	int SectorOffset();				// セクタオフセット取得
	void SetSectorOffset( int );	// セクタオフセット設定
	DWORD RelTime();				// インデックスホールからの時間(us)取得
	void SetRelTime( DWORD );		// インデックスホールからの時間(us)設定
	
	bool IsUniqueData();			// 単一データセクタ?
};


// トラッククラス
class FDTRK {
private:
	BYTE fdtype;					// ディスクの種類(1D,2D,1DD,2DD,2HD)
	std::vector<FDSEC> sectors;		// セクタオブジェクト
	
	void SetSectorOffset();			// セクタオフセット設定

public:
	FDTRK();						// コンストラクタ
	~FDTRK();						// デストラクタ
	
	void SetType( BYTE );			// ディスクの種類設定
	
	int Sectors();					// セクタ数取得
	FDSEC& Sector( const int = 1 );	// セクタ取得
	FDSEC& AddSector( BYTE, BYTE, BYTE, BYTE, BYTE, BYTE = 0, BYTE = 0 );	// セクタ追加
	BYTE Gap3();					// GAP3長取得
	BYTE EstimateGap3Data();		// GAP3データを推定する
	int SpritSector( int, BYTE );	// セクタ分割数を求める(ミックスセクタフォーマット時)
	
	bool Format( BYTE, BYTE, BYTE, BYTE, BYTE );	// フォーマット
	void Unformat();				// アンフォーマット
	
	BYTE ActualSectorSize( int );	// 実際に有効と思われるセクタサイズ取得(DATA CRC ERRORセクタの場合)
};


// フロッピーディスクイメージクラス
class FDimg {
protected:
	std::string title;				// ディスクの名前
	BYTE fdtype;					// ディスクの種類(1D,2D,1DD,2DD,2HD)
	bool protect;					// ライトプロテクト
	std::vector<FDTRK> track;		// トラックテーブル
	
	virtual bool ReadFileTrack ( std::fstream&, int, FDTRK& ) = 0;	// トラックデータ読込み
	virtual bool WriteFileTrack( std::fstream&, int, FDTRK& ) = 0;	// トラックデータ書込み
	
	bool IsSingleSide();			// 片面ディスク?

public:
	FDimg();						// コンストラクタ
	~FDimg();						// デストラクタ
	
	FDTRK& Track( BYTE, BYTE );		// トラック取得
	const std::string& Title() const;		// ディスクの名前取得
	void SetTitle( const std::string& );	// ディスクの名前設定
	const BYTE Type() const;		// ディスクの種類取得
	void SetType( BYTE );			// ディスクの種類設定
	const bool Protect() const;		// ライトプロテクト取得
	void SetProtect( bool );		// ライトプロテクト設定
	
	void Unformat();				// アンフォーマット
	
	virtual bool ReadFromFile( const std::string& ) = 0;			// イメージファイルから読込み
	virtual bool WriteToFile ( const std::string&, BYTE = 0 ) = 0;	// イメージファイルに書込み
};


// D88ディスクイメージクラス
class D88DISK : public FDimg {
private:
	bool ReadFileTrack ( std::fstream&, int, FDTRK& );	// トラックデータ読込み
	bool WriteFileTrack( std::fstream&, int, FDTRK& );	// トラックデータ書込み

public:
	D88DISK();						// コンストラクタ
	~D88DISK();						// デストラクタ
	
	bool ReadFromFile( const std::string& );			// イメージファイルから読込み
	bool WriteToFile ( const std::string&, BYTE = 0 );	// イメージファイルに書込み
};


#endif
